import { IsArray, IsEnum, IsNumber, IsOptional, IsString, ValidateNested } from "class-validator";
import { Type } from "class-transformer";

export class Gravity {
  @IsNumber()
  x: number;
  @IsNumber()
  y: number;
  @IsNumber()
  z: number;
}

export class Pose {
  @IsNumber()
  x: number;
  @IsNumber()
  y: number;
  @IsNumber()
  z: number;
  @IsNumber()
  roll: number;
  @IsNumber()
  pitch: number;
  @IsNumber()
  yaw: number;
}

export class Position {
  @IsNumber()
  x: number;
  @IsNumber()
  y: number;
  @IsNumber()
  z: number;
}

export enum InstanceType {
  RGB_CAMERA = "rgb_camera",
  SCENE_SIMPLE_OBJECT = "scene_simple_object",
}

abstract class CoreInstances {}

export class Instance extends CoreInstances {
  @IsEnum(InstanceType)
  instanceType: InstanceType;
  @Type(() => Position)
  position: Position;
  @IsArray()
  quaternion: number[];
  @IsOptional()
  @IsString()
  instanceAt: null | string = null;
}

export class SceneSimpleObject extends Instance {}

export class InstanceRgbCamera extends Instance {
  @IsString()
  cameraLink: string;
  @IsString()
  topicCameraInfo: string;
  @IsOptional()
  @IsString()
  topicDepth: string | null;
  @IsString()
  topicImage: string;
}
export class Asset {
  @IsString()
  name: string;
  @IsString()
  ixx: string;
  @IsString()
  ixy: string;
  @IsString()
  ixz: string;
  @IsString()
  iyy: string;
  @IsString()
  izz: string;
  @IsString()
  mass: string;
  @IsString()
  posX: string;
  @IsString()
  posY: string;
  @IsString()
  posZ: string;
  @IsString()
  eulerX: string;
  @IsString()
  eulerY: string;
  @IsString()
  eulerZ: string;
  @IsString()
  iyz: string;
  @IsString()
  meshPath: string;
  @IsString()
  friction: string;
  @IsString()
  centerMassX: string;
  @IsString()
  centerMassY: string;
  @IsString()
  centerMassZ: string;
  @IsArray()
  @IsOptional()
  actions: string[];
}

export class Physics {
  @IsString()
  engine_name: string;
  @Type(() => Gravity)
  gravity: Gravity;
}

export class RobossemblerAssets {
  @ValidateNested()
  @Type(() => Asset, {
    discriminator: {
      property: "type",
      subTypes: [
        { value: InstanceRgbCamera, name: InstanceType.RGB_CAMERA },
        { value: SceneSimpleObject, name: InstanceType.SCENE_SIMPLE_OBJECT },
      ],
    },
    keepDiscriminatorProperty: true,
  })
  assets: Asset[];

  @IsArray()
  @Type(() => Instance, {
    discriminator: {
      property: "type",
      subTypes: [
        { value: InstanceRgbCamera, name: InstanceType.RGB_CAMERA },
        { value: SceneSimpleObject, name: InstanceType.SCENE_SIMPLE_OBJECT },
      ],
    },
    keepDiscriminatorProperty: true,
  })
  instances: Instance[];

  @IsOptional()
  @ValidateNested()
  @Type(() => Physics)
  physics: Physics;

  convertLocalPathsToServerPaths(server_address: string): RobossemblerAssets {
    this.assets = this.assets.map((el) => {
      el.meshPath = server_address + el.meshPath;
      return el;
    });
    return this;
  }

  getAssetPath(assetName: string): string {
    const findElement = this.assets.find((el) => el.name === assetName);

    if (findElement === undefined) {
      throw new Error("RobossemblerAssets.getAssetPath not found asset by name:" + assetName);
    }
    return findElement.meshPath;
  }

  getAssetAtInstance(instanceAt: string): Asset {
    return this.assets.filter((el) => el.name === instanceAt)[0];
  }
}
