import {
  DirectionalLight,
  Object3D,
  PerspectiveCamera,
  Scene,
  WebGLRenderer,
  AmbientLight,
  Vector3,
  Mesh,
  Object3DEventMap,
  Box3,
  Sphere,
  LineBasicMaterial,
  Intersection,
  Raycaster,
  LineSegments,
  Vector2,
  Color,
  GridHelper,
  CameraHelper,
  Quaternion,
  MeshBasicMaterial,
  PlaneGeometry,
  BoxGeometry
} from "three";
import { TypedEvent } from "../helper/typed_event";
import { Result } from "../helper/result";
import { GLTFLoader, OrbitControls, TransformControls, OBJLoader, STLLoader, ColladaLoader } from "three-stdlib";
import {
  BaseSceneItemModel,
  CameraViewModel,
  StaticAssetItemModel,
} from "../../features/scene_manager/model/scene_assets";
import { SceneMode } from "../../features/scene_manager/model/scene_view";
import { throttle } from "../helper/throttle";
import { Asset, InstanceRgbCamera, RobossemblerAssets, SceneSimpleObject } from "../model/robossembler_assets";
 
export enum UserData {
  selectedObject = "selected_object",
  cameraInitialization = "camera_initialization",
  objectForMagnetism = "object_for_magnetism",
}

interface IEventDraggingChange {
  target: null;
  type: string;
  value: boolean;
}

interface IEmissiveCache {
  status: boolean;
  object3d: Object3D<Object3DEventMap>;
}

export class CoreThreeRepository extends TypedEvent<BaseSceneItemModel> {
  scene = new Scene();
  camera: PerspectiveCamera;
  webGlRender: WebGLRenderer;
  htmlCanvasRef: HTMLCanvasElement;
  objectEmissive = new Map<string, IEmissiveCache>();
  transformControls: TransformControls;
  orbitControls: OrbitControls;
  htmlSceneWidth: number;
  htmlSceneHeight: number;
  objLoader = new OBJLoader();
  glbLoader = new GLTFLoader();
  daeLoader = new ColladaLoader();
  stlLoader = new STLLoader();
  watcherSceneEditorObject: Function;

  constructor(htmlCanvasRef: HTMLCanvasElement, watcherSceneEditorObject: Function) {
    super();
    this.htmlSceneWidth = window.innerWidth;
    this.htmlSceneHeight = window.innerHeight;
    this.watcherSceneEditorObject = watcherSceneEditorObject;
    const renderer = new WebGLRenderer({
      canvas: htmlCanvasRef as HTMLCanvasElement,
      antialias: true,
      alpha: true,
    });
    const aspectCamera = this.htmlSceneWidth / this.htmlSceneHeight;
    this.camera = new PerspectiveCamera(800, aspectCamera, 0.1, 10000);
    this.camera.position.set(60, 20, 10);

    this.webGlRender = renderer;
    this.htmlCanvasRef = htmlCanvasRef;

    this.transformControls = new TransformControls(this.camera, htmlCanvasRef);

    this.scene.add(this.transformControls);
    this.orbitControls = new OrbitControls(this.camera, this.htmlCanvasRef);
    this.scene.background = new Color("black");

    this.init();
  }
  drawPoint(point: Vector3): Mesh<BoxGeometry, MeshBasicMaterial, Object3DEventMap> {
    var cube = new Mesh(new BoxGeometry(0.5, 0.5, 0.5), new MeshBasicMaterial({ color: 0x0095dd }));
    cube.position.add(point);
    this.scene.add(cube);
    return cube;
  }
  makeCube(inc: number, vector?: Vector3, color?: string, size?: number) {
    const cube = new Mesh(
      new BoxGeometry(size ?? 10, size ?? 10, size ?? 10),
      new MeshBasicMaterial({ color: color ?? 0x0095dd, transparent: true, opacity: 0.5 })
    );
    cube.userData[UserData.objectForMagnetism] = true;
    cube.userData[UserData.selectedObject] = true;
    cube.name = "cube" + String(inc);

    if (vector) {
      cube.position.copy(vector);
    }
    this.scene.add(cube);
  }
  makePoint(vector?: Vector3, color?: string, size?: number) {
    const cube = new Mesh(
      new BoxGeometry(size ?? 10, size ?? 10, size ?? 10),
      new MeshBasicMaterial({ color: color ?? 0x0095dd, transparent: true, opacity: 0.5 })
    );
    cube.userData[UserData.objectForMagnetism] = true;
    cube.userData[UserData.selectedObject] = true;

    if (vector) {
      cube.position.copy(vector);
    }
    this.scene.add(cube);
  }
  deleteSceneItem(item: BaseSceneItemModel) {
    const updateScene = this.scene;
    updateScene.children = item.deleteToScene(updateScene);
  }

  loadInstances(robossemblerAssets: RobossemblerAssets) {
    robossemblerAssets.instances.forEach(async (el) => {
      if (el instanceof InstanceRgbCamera) {
        const cameraModel = CameraViewModel.fromInstanceRgbCamera(el);
        cameraModel.mapPerspectiveCamera(this.htmlSceneWidth, this.htmlSceneHeight).forEach((el) => this.scene.add(el));
        this.emit(cameraModel);
      }
      if (el instanceof SceneSimpleObject) {
        const asset = robossemblerAssets.getAssetAtInstance(el.instanceAt as string);
        this.loader(
          asset.meshPath,
          () => {},
          asset.name,
          new Vector3(el.position.x, el.position.y, el.position.z),
          new Quaternion(el.quaternion[0], el.quaternion[1], el.quaternion[2], el.quaternion[3])
        );
      }
    });
  }
  loadInstance(asset: Asset, loadCallback?: Function) {
    this.loader(
      asset.meshPath,
      loadCallback ? loadCallback : () => {},
      asset.name,
      new Vector3(Number(asset.posX), Number(asset.posY), Number(asset.posZ)),
      new Quaternion(0, 0, 0, 0)
    );
  }

  setTransformMode(mode?: SceneMode) {
    switch (mode) {
      case undefined:
        this.transformControls.detach();
        this.transformControls.dispose();
        break;
      case SceneMode.MOVING:
        this.transformControls.setMode("translate");
        break;
      case SceneMode.ROTATE:
        this.transformControls.setMode("rotate");
        break;
    }
  }

  addSceneCamera(cameraModel: CameraViewModel) {
    cameraModel.mapPerspectiveCamera(this.htmlSceneWidth, this.htmlSceneHeight).forEach((el) => this.scene.add(el));
  }

  disposeTransformControlsMode() {
    this.transformControls.detach();
  }

  setRayCastAndGetFirstObjectName(vector: Vector2): Result<void, string> {
    this.scene.add(this.transformControls);
    const raycaster = new Raycaster();
    raycaster.setFromCamera(vector, this.camera);
    const intersects = raycaster.intersectObjects(this.scene.children);
    if (intersects.length > 0) {
      return Result.ok(intersects[0].object.name);
    }

    return Result.error(undefined);
  }

  setTransformControlsAttach(object: Object3D<Object3DEventMap>) {
    if (object instanceof CameraHelper) {
      this.transformControls.attach(object.camera);
      return;
    }
    this.transformControls.attach(object);
  }

  setRayCastAndGetFirstObject(vector: Vector2): Result<void, Object3D<Object3DEventMap>> {
    try {
      const result = this.setRayCast(vector);
      return result.fold(
        (intersects) => {
          const result = intersects.find((element) => element.object.userData[UserData.selectedObject] !== undefined);
          if (result === undefined) {
            return Result.error(undefined);
          }
          return Result.ok(result.object);
        },
        (_error) => {
          return Result.error(undefined);
        }
      );
    } catch (error) {
      return Result.error(undefined);
    }
  }

  setRayCast(vector: Vector2): Result<void, Intersection<Object3D<Object3DEventMap>>[]> {
    const raycaster = new Raycaster();
    raycaster.setFromCamera(vector, this.camera);
    const intersects = raycaster.intersectObjects(this.scene.children);
    if (intersects.length > 0) {
      return Result.ok(intersects);
    }

    return Result.error(undefined);
  }

  setRayCastAndGetFirstObjectAndPointToObject(vector: Vector2): Result<void, Vector3> {
    this.setRayCast(vector).map((intersects) => {
      if (intersects.length > 0) {
        return Result.ok(intersects[0].point);
      }
    });
    return Result.error(undefined);
  }

  light() {
    const ambientLight = new AmbientLight(0xffffff, 0.7);
    this.scene.add(ambientLight);

    this.scene.add(new AmbientLight(0x222222));

    this.scene.add(new DirectionalLight(0xffffff, 1));
  }

  addListeners() {
    window.addEventListener(
      "resize",
      () => {
        this.camera.aspect = this.htmlSceneWidth / this.htmlSceneHeight;
        this.camera.updateProjectionMatrix();

        this.webGlRender.setSize(this.htmlSceneWidth, this.htmlSceneHeight);
      },
      false
    );

    this.transformControls.addEventListener("dragging-changed", (event) => {
      const e = event as unknown as IEventDraggingChange;
      this.orbitControls.enabled = !e.value;
    });
    this.transformControls.addEventListener("objectChange", (event) => {
      //@ts-expect-error
      const sceneObject = event.target.object;
      //TODO:(IDONTSUDO) Trotting doesn't work, need to figure out why
      const fn = () => this.watcherSceneEditorObject(sceneObject);
      const [throttleFn] = throttle(fn, 1000);
      throttleFn();
    });
  }

  init() {
    this.light();
    this.addListeners();
    const floor = new GridHelper(100, 100, 0x888888, 0x444444);
    floor.userData = {};
    floor.userData[UserData.cameraInitialization] = true;
    this.scene.add(floor);

    const planeMesh = new Mesh(
      new PlaneGeometry(10, 10, 10, 10),
      new MeshBasicMaterial({ color: 0x808080, wireframe: true })
    );
    planeMesh.userData[UserData.selectedObject] = true;
    planeMesh.userData[UserData.objectForMagnetism] = true;
    planeMesh.rotation.x = -Math.PI / 2;
    this.makeCube(1);
    this.makeCube(2, new Vector3(20, 0, 10), "yellow");

    this.scene.add(planeMesh);
  }

  render() {
    this.webGlRender.setSize(this.htmlSceneWidth, this.htmlSceneHeight);
    this.webGlRender.setAnimationLoop(() => {
      this.webGlRender.render(this.scene, this.camera);
    });
  }

  getAllSceneModels(): BaseSceneItemModel[] {
    return this.getAllSceneNameModels().map(
      (name) =>
        new StaticAssetItemModel(name, this.getObjectsAtName(name).position, this.getObjectsAtName(name).quaternion)
    );
  }

  getAllSceneNameModels(): string[] {
    return this.scene.children.filter((el) => el.name !== "").map((el) => el.name);
  }

  getObjectsAtName(name: string): Object3D<Object3DEventMap> {
    return this.scene.children.filter((el) => el.name === name)[0];
  }

  loader(url: string, callBack: Function, name: string, position?: Vector3, quaternion?: Quaternion) {
    const ext = url.split(/\./g).pop()!.toLowerCase();

    switch (ext) {
      case "gltf":
      case "glb":
        this.glbLoader.load(
          url,
          (result) => {},
          (err) => {}
        );
        break;
      case "obj":
        this.objLoader.load(
          url,
          (result) => {
            result.userData[UserData.selectedObject] = true;
            result.children.forEach((el) => {
              el.userData[UserData.selectedObject] = true;
              el.name = name;

              if (position) el.position.copy(position);
              if (quaternion) el.quaternion.copy(quaternion);

              this.emit(new StaticAssetItemModel(el.name, el.position, el.quaternion));
              this.scene.add(el);
            });
          },
          (err) => {}
        );
        break;
      case "dae":
        this.daeLoader.load(
          url,
          (result) => {},
          (err) => {}
        );
        break;
      case "stl":
        this.stlLoader.load(
          url,
          (result) => {},

          (err) => {}
        );
        break;
    }
  }

  fitCameraToCenteredObject(objects: string[], offset = 4) {
    const boundingBox = new Box3().setFromPoints(
      objects.map((el) => this.getObjectsAtName(el)).map((el) => el.position)
    );

    var size = new Vector3();
    boundingBox.getSize(size);

    const fov = this.camera.fov * (Math.PI / 180);
    const fovh = 2 * Math.atan(Math.tan(fov / 2) * this.camera.aspect);
    let dx = size.z / 2 + Math.abs(size.x / 2 / Math.tan(fovh / 2));
    let dy = size.z / 2 + Math.abs(size.y / 2 / Math.tan(fov / 2));
    let cameraZ = Math.max(dx, dy);

    if (offset !== undefined && offset !== 0) cameraZ *= offset;

    this.camera.position.set(0, 0, cameraZ);

    const minZ = boundingBox.min.z;
    const cameraToFarEdge = minZ < 0 ? -minZ + cameraZ : cameraZ - minZ;

    this.camera.far = cameraToFarEdge * 3;
    this.camera.updateProjectionMatrix();
    let orbitControls = new OrbitControls(this.camera, this.htmlCanvasRef);

    orbitControls.maxDistance = cameraToFarEdge * 2;
    this.orbitControls = orbitControls;
  }

  switchObjectEmissive(name: string) {
    const mesh = this.getObjectsAtName(name);
    const result = this.objectEmissive.get(mesh.name);
    if (result?.status) {
      this.scene.remove(mesh);
      this.scene.add(result.object3d);
      this.objectEmissive.set(mesh.name, {
        status: false,
        object3d: mesh,
      });
    } else {
      this.objectEmissive.set(mesh.name, {
        status: true,
        object3d: mesh,
      });

      if (mesh instanceof Mesh) {
        const newMesh = new LineSegments(mesh.geometry, new LineBasicMaterial({ color: 0x000000 }));
        newMesh.name = mesh.name;
        newMesh.position.copy(mesh.position);

        this.scene.remove(mesh);
        this.scene.add(newMesh);
      }
    }
  }

  fitSelectedObjectToScreen(objects: string[]) {
    //https://stackoverflow.com/questions/14614252/how-to-fit-camera-to-object
    let boundBox = new Box3().setFromPoints(objects.map((el) => this.getObjectsAtName(el)).map((el) => el.position));
    let boundSphere = boundBox.getBoundingSphere(new Sphere());
    let vFoV = this.camera.getEffectiveFOV();
    let hFoV = this.camera.fov * this.camera.aspect;
    let FoV = Math.min(vFoV, hFoV);
    let FoV2 = FoV / 2;
    let dir = new Vector3();
    this.camera.getWorldDirection(dir);
    let bsWorld = boundSphere.center.clone();
    let th = (FoV2 * Math.PI) / 180.0;
    let sina = Math.sin(th);
    let R = boundSphere.radius;
    let FL = R / sina;
    let cameraDir = new Vector3();
    let cameraOffs = cameraDir.clone();
    cameraOffs.multiplyScalar(-FL);

    let newCameraPos = bsWorld.clone().add(cameraOffs);

    this.camera.position.copy(newCameraPos);
    this.camera.lookAt(bsWorld);
    this.orbitControls = new OrbitControls(this.camera, this.htmlCanvasRef);
  }
}
