import { createRoot } from "react-dom/client";
import { NodeEditor, GetSchemes, ClassicPreset } from "rete";
import { AreaPlugin, AreaExtensions } from "rete-area-plugin";
import { ConnectionPlugin, Presets as ConnectionPresets } from "rete-connection-plugin";
import { ReactPlugin, Presets, ReactArea2D } from "rete-react-plugin";
import { CustomSocket } from "./custom_socket";
import { CustomConnection } from "./custom_connection";
import { addCustomBackground } from "./custom_background";
import { behaviorTreeBuilderStore } from "../../behavior_tree_builder_screen";
import { SequenceNode } from "./nodes/controls_node";
import { ReteObserver } from "../../../model/editor_view";

export type Schemes = GetSchemes<ClassicPreset.Node, ClassicPreset.Connection<ClassicPreset.Node, ClassicPreset.Node>>;
export type AreaExtra = ReactArea2D<Schemes>;

export async function createEditor(container: HTMLElement) {
  const socket = new ClassicPreset.Socket("socket");
  const observer = new ReteObserver();

  behaviorTreeBuilderStore.reteNode = observer;

  behaviorTreeBuilderStore.btNodeView.on(async (event) => {
    setTimeout(async () => {
      const node = new ClassicPreset.Node(event.name);

      const { x, y } = areaContainer.area.pointer;

      node.addOutput("a", new ClassicPreset.Output(socket));
      node.addInput("a", new ClassicPreset.Input(socket));
      await editor.addNode(node);
      await areaContainer.translate(node.id, { x, y });
    }, 100);
  });

  observer.on(() => {
    behaviorTreeBuilderStore.bt(editor, areaContainer);
  });

  const editor = new NodeEditor<Schemes>();
  const areaContainer = new AreaPlugin<Schemes, AreaExtra>(container);
  const connection = new ConnectionPlugin<Schemes, AreaExtra>();
  const render = new ReactPlugin<Schemes, AreaExtra>({ createRoot });

  AreaExtensions.selectableNodes(areaContainer, AreaExtensions.selector(), {
    accumulating: AreaExtensions.accumulateOnCtrl(),
  });

  render.addPreset(
    Presets.classic.setup({
      customize: {
        node(context) {
          return SequenceNode;
        },
        socket(_context) {
          return CustomSocket;
        },
        connection(_context) {
          return CustomConnection;
        },
      },
    })
  );

  connection.addPreset(ConnectionPresets.classic.setup());

  addCustomBackground(areaContainer);

  editor.use(areaContainer);
  areaContainer.use(connection);
  areaContainer.use(render);

  AreaExtensions.simpleNodesOrder(areaContainer);

  for await (const el of behaviorTreeBuilderStore.nodes) {
    const node = new ClassicPreset.Node(el.label);
    node.id = el.id;
    el.outputs.forEach((outputName) => {
      node.addOutput(outputName, new ClassicPreset.Output(socket));
    });
    el.inputs.forEach((inputName) => {
      node.addInput(inputName, new ClassicPreset.Input(socket));
    });

    await editor.addNode(node);
    await areaContainer.translate(node.id, el.position);
  }

  behaviorTreeBuilderStore.nodes.forEach(async (el) => {
    if (el.connectTo) {
      const nodeConnectTo = editor.getNode(el.connectTo!);
      const nodeConnect = editor.getNode(el.id);
      await editor.addConnection(new ClassicPreset.Connection(nodeConnectTo, "a", nodeConnect, "a"));
    }
  });

  setTimeout(() => {
    AreaExtensions.zoomAt(areaContainer, editor.getNodes());
  }, 100);

  return {
    destroy: () => areaContainer.destroy(),
  };
}
