import { CameraHelper, Object3D, PerspectiveCamera, Quaternion, Scene, Vector3 } from "three";
import { v4 as uuidv4 } from "uuid";
import { UserData } from "../../../core/repository/core_three_repository";
import {
  Asset,
  Instance,
  InstanceRgbCamera,
  InstanceType,
  SceneSimpleObject,
} from "../../../core/model/robossembler_assets";

export enum RobossemblerFiles {
  robossemblerAssets = "robossembler_assets.json",
}

export enum SceneModelsType {
  ASSET,
}

export abstract class BaseSceneItemModel {
  id: string;
  name: string;
  position: Vector3;
  quaternion: Quaternion;

  constructor(name: string) {
    this.id = uuidv4();
    this.name = name;
  }

  abstract toInstance(): Instance;

  abstract deleteToScene(scene: Scene): Object3D[];

  toQuaternionArray() {
    return [this.quaternion.x, this.quaternion.y, this.quaternion.z, this.quaternion.w];
  }
}

export class StaticAssetItemModel extends BaseSceneItemModel {
  name: string;
  type = SceneModelsType.ASSET;

  constructor(name: string, position: Vector3, quaternion: Quaternion) {
    super(name);
    this.name = name;
    this.position = position;
    this.quaternion = quaternion;
  }

  static fromSceneSimpleObjectAndAsset(sceneSimpleObject: SceneSimpleObject, asset: Asset) {
    const { x, y, z } = sceneSimpleObject.position;
    const quaternion = sceneSimpleObject.quaternion;
    return new StaticAssetItemModel(
      asset.name,
      new Vector3(x, y, z),
      new Quaternion(quaternion[0], quaternion[1], quaternion[2], quaternion[3])
    );
  }

  toInstance(): Instance {
    const instance = new Instance();
    instance.instanceType = InstanceType.SCENE_SIMPLE_OBJECT;
    instance.position = this.position;
    instance.instanceAt = this.name;
    instance.quaternion = this.toQuaternionArray();

    return instance;
  }

  deleteToScene(scene: Scene): Object3D[] {
    return scene.children.filter((el) => el.name !== this.name);
  }
}

export interface ICoreInstance {
  instanceAt: null | string;
  type: string;
  position: Vector3;
  quaternion: Quaternion;
}

export interface ICameraInstance extends ICoreInstance {
  cameraLink: string;
  topicImage: string;
  topicCameraInfo: string;
  topicDepth: string | null;
}

export class CameraViewModel extends BaseSceneItemModel {
  cameraLink: string;
  topicImage: string;
  topicCameraInfo: string;
  topicDepth: null | string;

  constructor(cameraLink: string, topicImage: string, topicCameraInfo: string, topicDepth: string | null) {
    super(cameraLink);
    this.cameraLink = cameraLink;
    this.topicImage = topicImage;
    this.topicCameraInfo = topicCameraInfo;
    this.topicDepth = topicDepth;
  }

  static fromInstanceRgbCamera(instanceRgbCamera: InstanceRgbCamera) {
    const { cameraLink, topicImage, topicCameraInfo, topicDepth, position, quaternion } = instanceRgbCamera;
    const instance = new CameraViewModel(cameraLink, topicImage, topicCameraInfo, topicDepth);
    const { x, y, z } = position;

    instance.position = new Vector3(x, y, z);
    instance.quaternion = new Quaternion(quaternion[0], quaternion[1], quaternion[2], quaternion[3]);

    return instance;
  }

  deleteToScene(scene: Scene): Object3D[] {
    return scene.children.filter((el) => {
      if (el.name === this.name) {
        return false;
      }
      if (el.name === this.cameraLink + "camera_helper") {
        return false;
      }
      return true;
    });
  }

  toInstance(): InstanceRgbCamera {
    return {
      instanceType: InstanceType.RGB_CAMERA,
      position: this.position,
      quaternion: this.toQuaternionArray(),
      instanceAt: null,
      cameraLink: this.cameraLink,
      topicCameraInfo: this.topicCameraInfo,
      topicDepth: this.topicDepth,
      topicImage: this.topicImage,
    };
  }

  validate(cameraLinksNames: string[]) {
    if (cameraLinksNames.filter((el) => this.cameraLink === el).isNotEmpty()) {
      return { cameraLink: "the name for the camera is not unique" };
    }
    if (this.cameraLink.isEmpty()) {
      return { cameraLink: "is empty" };
    }
    if (this.topicImage.isEmpty()) {
      return { topicImage: "is empty" };
    }
    if (this.topicCameraInfo.isEmpty()) {
      return { topicCameraInfo: "is empty" };
    }
  }

  mapPerspectiveCamera(htmlSceneWidth: number, htmlSceneHeight: number) {
    const perspectiveCamera = new PerspectiveCamera(48, htmlSceneWidth / htmlSceneHeight, 7.1, 28.5);
    perspectiveCamera.position.copy(this.position);
    perspectiveCamera.quaternion.copy(this.quaternion);
    perspectiveCamera.name = this.cameraLink;
    const cameraHelper = new CameraHelper(perspectiveCamera);
    perspectiveCamera.userData[UserData.selectedObject] = true;
    cameraHelper.userData[UserData.selectedObject] = true;
    cameraHelper.name = this.cameraLink + "camera_helper";
    return [cameraHelper, perspectiveCamera];
  }

  static empty() {
    return new CameraViewModel("", "", "", "");
  }
}
