import makeAutoObservable from "mobx-store-inheritance";
import { CoreThreeRepository, UserData } from "../../../core/repository/core_three_repository";
import { Object3D, Vector2, Vector3 } from "three";
import { HttpError } from "../../../core/repository/http_repository";
import { UiErrorState } from "../../../core/store/base_store";
import { UiBaseError } from "../../../core/model/ui_base_error";
import { SceneMenu, SceneMode } from "../model/scene_view";
import { BaseSceneItemModel, CameraViewModel, RobossemblerFiles, StaticAssetItemModel } from "../model/scene_assets";
import { SceneHttpRepository } from "../data/scene_repository";
import { message } from "antd";
import { RobossemblerAssets } from "../../../core/model/robossembler_assets";

export class SceneMangerStore extends UiErrorState<HttpError> {
  sceneMode: SceneMode;
  sceneMenu: SceneMenu;
  isVisibleSaveButton: boolean = false;
  coreThereRepository: null | CoreThreeRepository = null;
  sceneHttpRepository: SceneHttpRepository;
  sceneModels: BaseSceneItemModel[] = [];
  isSceneMenuShow = false;
  robossemblerAssets?: RobossemblerAssets;
  objectForMagnetism: string;
  objectMagnetism: string;

  constructor() {
    super();
    makeAutoObservable(this);
    this.sceneHttpRepository = new SceneHttpRepository();
    this.sceneMode = SceneMode.EMPTY;
    this.sceneMenu = SceneMenu.empty();
  }

  onTapSave(): void {
    this.robossemblerAssets!.instances = [];
    this.sceneModels.forEach((el) => this.robossemblerAssets?.instances.push(el.toInstance()));
    this.httpHelper(this.sceneHttpRepository.saveScene(this.robossemblerAssets as RobossemblerAssets));
    this.isVisibleSaveButton = false;
  }

  deleteSceneItem(item: BaseSceneItemModel) {
    const itm = this.sceneModels.filter((el) => el.id === item.id);
    this.coreThereRepository!.deleteSceneItem(itm[0]);
    this.sceneModels = this.sceneModels.filter((el) => el.name !== item.name);
    this.visibleSaveButton();
  }

  visibleSaveButton() {
    this.isVisibleSaveButton = true;
  }

  addNewCamera(model: CameraViewModel) {
    model.position = this.coreThereRepository!.camera.position;
    model.quaternion = this.coreThereRepository!.camera.quaternion;
    this.sceneModels.push(model);
    this.coreThereRepository?.addSceneCamera(model);
    this.visibleSaveButton();
  }

  getCameraLinkNames(): string[] {
    return this.sceneModels.map((el) => {
      if (el instanceof CameraViewModel) {
        return el.cameraLink;
      }
      return "";
    });
  }

  loaderWatcher() {}

  loadSceneRobossemblerAsset(name: string) {
    try {
      const assetPath = this.robossemblerAssets?.getAssetPath(name) as string;
      this.coreThereRepository?.loader(assetPath, this.loaderWatcher, name);
      this.visibleSaveButton();
    } catch (error) {
      message.error(String(error));
    }
  }

  isRenderedAsset(name: string): boolean {
    return this.sceneModels
      .filter((el) => {
        if (el instanceof StaticAssetItemModel) {
          return el.name === name;
        }
        return false;
      })
      .isNotEmpty();
  }

  hiddenMenu() {
    this.isSceneMenuShow = false;
  }

  setSceneMode = (mode: SceneMode) => {
    if (this.sceneMode === undefined || this.sceneMode !== mode) {
      this.sceneMode = mode;
    } else if (this.sceneMode === mode) {
      this.sceneMode = SceneMode.EMPTY;
    }
    this.coreThereRepository?.setTransformMode(this.sceneMode);
    this.sceneModeWatcher();
  };

  sceneModeWatcher() {}

  async init(): Promise<any> {}

  errorHandingStrategy = (error: HttpError) => {
    if (error.status === 404) {
      this.errors.push(new UiBaseError(`${RobossemblerFiles.robossemblerAssets} not found to project`));
    }
  };

  async loadScene(canvasRef: HTMLCanvasElement) {
    this.loadWebGl(canvasRef);
    await this.mapOk<RobossemblerAssets>("robossemblerAssets", this.sceneHttpRepository.getRobossemblerAssets());
    if (this.robossemblerAssets) {
      this.coreThereRepository?.loadInstances(this.robossemblerAssets);
    }
  }

  loadWebGl(canvasRef: HTMLCanvasElement): void {
    this.coreThereRepository = new CoreThreeRepository(canvasRef as HTMLCanvasElement, this.watcherSceneEditorObject);
    this.coreThereRepository.on(this.watcherThereObjects);
    this.coreThereRepository.render();
    this.sceneModels = this.coreThereRepository.getAllSceneModels();

    window.addEventListener("click", (event) => this.clickLister(event));
    window.addEventListener("mousedown", (e) => this.sceneContextMenu(e));
  }

  clickLister(event: MouseEvent) {
    const vector = new Vector2();
    vector.x = (event.clientX / window.innerWidth) * 2 - 1;
    vector.y = -(event.clientY / window.innerHeight) * 2 + 1;
    if (this.sceneMode === SceneMode.EMPTY) {
      return;
    }

    if (this.sceneMode === SceneMode.MOVING || this.sceneMode === SceneMode.ROTATE) {
      this.transformContollsCall(vector);
    }
  }

  sceneContextMenu(e: MouseEvent): void {
    if (e.button === 2) {
      this.isSceneMenuShow = true;
      this.sceneMenu.x = e.clientX;
      this.sceneMenu.y = e.clientY;
    }
  }

  watcherThereObjects = (sceneItemModel: BaseSceneItemModel): void => {
    this.sceneModels.push(sceneItemModel);
  };

  watcherSceneEditorObject = (mesh: Object3D) => {
    this.sceneModels = this.sceneModels.map((el) => {
      if (el.name === mesh.name) {
        el.position = mesh.position;
        el.quaternion = mesh.quaternion;
        return el;
      }
      return el;
    });
    this.visibleSaveButton();
  };

  transformContollsCall = (vector: Vector2) => {
    this.coreThereRepository?.setRayCastAndGetFirstObject(vector).fold(
      (success) => this.coreThereRepository?.setTransformControlsAttach(success),
      (_error) => this.coreThereRepository?.disposeTransformControlsMode()
    );
  };

  dispose() {
    window.removeEventListener("click", this.clickLister);
    window.removeEventListener("mousedown", (e) => this.sceneContextMenu(e));
  }
}
